import os
from json import dumps, loads
from datetime import datetime
from collections import defaultdict

from testyoke.models import TestSuite, TestCase


HISTORY_FILE = os.path.join(
    os.environ.get("HOME", ''), '.testyoke/'
)

RESULTS_FILE = 'results'


def history_file(project, filename):
    # TODO """ clean up dir concat """
    return HISTORY_FILE + project + "/" + filename


def bootstrap(project):
    # TODO check if dir exists
    try:
        if not os.path.exists(HISTORY_FILE):
            os.mkdir(HISTORY_FILE)
        if not os.path.exists(HISTORY_FILE + project):
            os.mkdir(HISTORY_FILE + project)

        results = history_file(project, RESULTS_FILE)
        if not os.path.exists(results):
            open(results, 'w').write('')

    except FileExistsError as err:
        raise Exception("Error creating info dir %s" % str(err)) 
        pass


def save(suites, project="default", sha=None):
    bootstrap(project)
    with open(history_file(project, RESULTS_FILE), 'a') as fp:
        for suite in suites:
            for case in suite.cases:
                to_write = case.to_dict
                to_write['suite'] = suite.to_dict
                to_write['request'] = {
                    'timestamp': datetime.now().isoformat(),
                }

                if sha is not None:
                    to_write['request']['sha'] = sha

                fp.write(dumps(to_write) + "\n")

def all_load(project):
    bootstrap(project)
    with open(history_file(project, RESULTS_FILE), 'r') as fp:
        for line in fp:
            data = loads(line)
            request = data['request']
            suite = TestSuite(**data.get('suite', {}))
            case = TestCase(sha=request.get('sha'), **data)

            yield (case, suite, request)

    
def load(project, **options):
    results = defaultdict(dict)

    # load
    case_suites = all_load(project)
    # filter
    if options.get('flaky'):
        case_suites = only_flaky(case_suites)

    for (case, suite, request) in case_suites:
        # TODO: this is screaming for an Aggregate class
        if results[suite.name].get(case.name) is None: 
            results[suite.name][case.name] = {
                'runs': 0, 'fails': 0, 
            }

        results[suite.name][case.name]['runs'] += 1
        if case.did_fail:
            results[suite.name][case.name]['fails'] += 1
        else:
            results[suite.name][case.name]['latest_pass'] = {
                'timestamp': request.get('timestamp'),
                'vc-sha': request.get('sha')
            }

        runs = float(results[suite.name][case.name]['runs'])
        fails = float(results[suite.name][case.name]['fails'])
        results[suite.name][case.name]['pass_rate'] = round((
            (runs-fails)/runs
        ) * 100, 2)

    return results


def only_flaky(case_suites):
    shas_pass = {}

    flaky = []

    for (case, suite, request) in case_suites:
        if case.did_fail:
            if shas_pass.get(case.sha) == True:
                flaky.append((case, suite, request))
            shas_pass[case.sha] = False

        else:
            if shas_pass.get(case.sha) == False:
                flaky.append((case, suite, request))
            shas_pass[case.sha] = True

    return flaky    


def filter(cases, **options):
    """ filter results according to `options` """
    cases = [case for case in cases]
    if options.get('flaky') is not None:
        cases = only_flaky(cases)

    return cases



from testyoke.server import get_parser


class TestServer:
    def test_get_parser_junit(self):
        content_type = 'application/xml+junit'
        parser = get_parser(content_type)

        assert callable(parser.parse)

    def test_get_parser_default(self):
        content_type = None
        parser = get_parser(content_type)

        assert callable(parser.parse)




from testyoke.models import TestSuite, TestCase
from testyoke.persist import file as persist


class TestFile:

    def test_saves_and_retrieves(self):
        suite = TestSuite()
        suite.cases = [
            TestCase(),
            TestCase()
        ]

        persist.save([suite], 'test-suite')
        # TODO: serialize testcase \n




import xml.etree.ElementTree as ET
from testyoke.models import TestSuite, TestCase

def parse(filename, **options):
    """ parses junit.xml, ALWAYS returns suites """
    root = ET.fromstring(filename)

    suites = []

    if root.tag == 'testsuite':
        root = [root]

    for suite_xml in root:
        suite = TestSuite(**suite_xml.attrib)

        for testcase in suite_xml:
            case = TestCase(**testcase.attrib)
            case.sha = options.get('sha')

            failure = testcase.find('failure')
            case.message = failure
            if failure is not None:
                case.message = failure.attrib['message']
                case.trace = failure.text

            suite.cases.append(case)

        suites.append(suite)
        # can impl TestSuite here.
    
    return suites



from testyoke.formats.xml import junit as parser


JUNIT_FIXTURE = 'test/fixtures/junit.xml'

class TestJunit:

    def test_read(self):
        # TODO missing <testsuites>
        xml = open(JUNIT_FIXTURE, 'r').read()
        suites = parser.parse(xml)
       
        assert suites[0].name == 'pytest' 
        assert suites[0].tests == 11

        assert suites[0].cases[0].name == 'test_benchmark'
        assert suites[0].cases[0].line == 6
        assert suites[0].cases[0].file == "tests/rsci/test_telemetry.py" 

    def test_reads_errors(self):
        # TODO missing <testsuites>
        xml = open("test-results/junit-20191020111034.xml", 'r').read()
        suites = parser.parse(xml)
       
        assert suites[0].name == 'pytest' 
        assert suites[0].tests == 6
        assert suites[0].failures == 1

        assert suites[0].cases[0].name == 'test_receives_report'

        assert suites[0].cases[3].name == 'test_read'
        assert suites[0].cases[3].did_fail



from testyoke import api
from testyoke.models import TestSuite, TestCase

PROJECT = 'default'

class TestApi:
    def test_receives_report(self):
        suites = [
            TestSuite()
        ]

        result = api.receive_report(suites, PROJECT)
        assert result
        
    def test_get_stats(self):
        result = api.get_stats(PROJECT)

        assert result['runs'] == {}
        



#
# EXPERIMENTAL
#
# functional passing of cases and suites
# outputs insights
#
# if it takes more than one pass to label a test, that's
# ok, because they are the same test.
#
# `persist` needs to eventually have some snapshot of tests
#
# does this work?
#
# result_stream -> analytics_mark_up -> save, snapshot
#
# can get stream of cases from service or persist
# TODO need to pull stuff out of persist.file
#
#
# Need CQRS pattern. can warm up from disk, or not,
# depends on the analysis' needs.
# expect this dir to break down into parts
# TESTS
#
from .sha import Analyzer as ShaAnalyzer

ANALYZERS = {
    'sha': ShaAnalyzer()
}


def analyze_test(case, suite, **options):
    # TODO needs project
    """ 
    pass test cases through to mark up analytical data

    case - 
    suite -
    environ - ?
    **options
    """

    mark_flaky(case, suite, **options)
    mark_regressive(case, suite, **options)

    # for each analyzer
    for k, analyzer in ANALYZERS.items():
        analyzer.analyze(case, suite, **options)


def get_sha(sha):
    """ shouldn't live here. """
    return ANALYZERS['sha'].get(sha)


def get_shas():
    """ shouldn't live here. """
    return ANALYZERS['sha'].all()

#
# TODO belongs in test
#

def mark_flaky(case, suite, **options):
    """ 
    flaky - passes and fails on same SHA.
    check if is_flaky, mark test 
    """
    if False:
        case.is_flaky = True


def mark_regressive(case, suite, **options):
    """
    regression - fails, is fixed, fails on new commit. not flaky.
    b
    """
    if False:
        case.is_regressive = True



#
# flask server for receiving data. statsd server?
#
import os
from datetime import datetime
from testyoke.persist import file as persist
from testyoke import analytics


def analyze(project, case, suite):
    analytics.analyze_test(case, suite)


def on_load(project):
    for case, suite, request in persist.all_load(project):
        analyze(project, case, suite)


def receive_report(suites, project=None, sha=None):
    persist.save(suites, project, sha)

    # TODO needs to happen here, and on load of file
    for suite in suites:
        for case in suite.cases:
            analyze(project, case, suite)

    return True

       
def get_stats(project, **options):
    return persist.load(project, **options)


def get_sha(project, sha):
    on_load('testharness')  # this will eventually have to be cached
    return analytics.get_sha(sha)


def get_shas(project):
    on_load('testharness') 
    return analytics.get_shas()




import json
from flask import Flask, request, Response
from testyoke.formats.xml import junit
from xml.etree.ElementTree import ParseError

from testyoke import api

app = Flask(__name__)

CONTENT_TYPES = {
    'application/xml+junit': junit
}

def get_parser(content_type):
    return CONTENT_TYPES.get(content_type, junit)

def headers():
    return {
        "Content-Type": "application/json"
    }

def params():
    return {
        'flaky': request.args.get('flaky') is not None
    }


@app.route('/projects/<string:project>/reports', methods=['POST'])
def receive_report(project):
    sha = request.headers.get('vc-sha', '').strip()

    try:
        suites = get_parser(
            request.headers['content-type']
        ).parse(request.get_data(), sha=sha)
    except ParseError:
        return '{"message": "report not parsable"}', 400

    api.receive_report(suites, project, sha)

    return "", 201


@app.route('/projects/<string:project>')
def stats_project(project):
    return Response(json.dumps(api.get_stats(project, **params())), headers=headers())


@app.route('/projects/<string:project>/<string:suite>')
def stats_suite(project, suite):
    case = api.get_stats(project, **params()).get(suite)
    return Response(json.dumps(case), headers=headers())


@app.route('/projects/<string:project>/<string:suite>/<string:test>')
def stats_test(project, suite, test):
    case = api.get_stats(project, **params()).get(suite).get(test)
    return Response(json.dumps(case), headers=headers())

#
# VC SHAS
#

@app.route('/projects/<string:project>/shas') 
def get_shas(project):
    """ missing project """
    return Response(json.dumps(api.get_shas(project)), headers=headers())


@app.route('/projects/<string:project>/shas/<string:sha>') 
def get_sha(project, sha):
    """ missing project """
    sha = api.get_sha(project, sha)

    # TODO GARBAGE
    response = sha.to_dict if sha is not None else {'message': 'Not Found'}
    return Response(json.dumps(response), headers=headers())

#
# welcome
#

@app.route('/')
def welcome():
    return 'testharness <a href="/projects/aproj/stats">aproj</a>'




import json
from http.client import HTTPConnection
from testyoke.models import ProjectState

class ClientException(Exception): pass

class Client:

    def __init__(self, project, hostname='localhost', port=7357):
        self.conn = HTTPConnection(hostname, port)
        self.project = project

    def sha(self, sha):
        self.conn.request("GET", "/projects/%s/shas/%s" % (self.project, sha))
        res = self.conn.getresponse()

        if res.status == 200:
            data = json.loads(res.read())
            return ProjectState(**data)
        else:
            raise ClientException("%s: %s" % (res.status, res.reason))
        
    # def case(self, case):
    # def report(self):
        
    def post(self, body, sha=None):
        headers = {
            'Content-Type': 'application/xml+junit',
            'vc-sha': sha
        }
        self.conn.request(
            "POST", 
            "/projects/%s/reports" % self.project, body, 
            headers
        )
        res = self.conn.getresponse()

        if res.status in [200, 201]:
            return True
        else:
            raise ClientException("%s: %s" % (res.status, res.reason))




#
# local in mem storage
# desigining this for one suite presently, doesn't scale
#
# need to hash by project, node, test, time
from datetime import datetime


class Analytics:
    def __init__(self):
        self.runs = 0
        self.fails = 0
        self.last_pass = None
        self.failures = []

    def add_outcomes(self, suites):
        failures = []

        for suite in suites:
            for case in suite.cases:
                self.runs = self.runs + 1

                if case.did_fail:
                    failures.append(case)
                    self.fails = self.fails + 1

        if not failures:
            self.last_pass = datetime.now()  # should be from request

        self.failures = failures

    @property
    def to_dict(self):
        return {
            'runs': self.runs
        }



from optparse import OptionParser
from os import environ
from glob import glob

from . import Client


def post_report(report, sha):
    with open(report, 'r') as f:
        report = f.read()
        client.post(report, sha=sha)

            
if __name__ == '__main__':
    parser = OptionParser()
    parser.add_option("-s", "--sha", dest="sha",
                  help="retrieve report for provided sha", metavar="VC SHA")
    parser.add_option("-p", "--project", dest="project",
                  help="name for this project", metavar="PROJECT")
    parser.add_option("-r", "--report", dest="report",
                  help="test report file to submit (Unix globbing*)", metavar="JUNIT.xml")

    opts, args = parser.parse_args()

    if not opts.project or (not opts.sha and not opts.report):
        parser.print_help()
        exit(0)

    client = Client(opts.project)

    if opts.sha and not opts.report:  # should this return after report uploaded?
        sha = client.sha(opts.sha)

        print()
        print("########################################################")
        print(f"# sha: {sha.sha.ljust(47)} #")
        print(f"# {repr(sha).ljust(53)}#")
        print("#                                                      #")
        print("########################################################")
        print()


    if opts.report:
        print(f"matching: {opts.report}", end=" ")
        for report_file in glob(opts.report):
            post_report(report_file, opts.sha)
            print('', end=".")

        print('done.')



import setuptools

from testyoke.version import __version__

with open('README.md', 'r') as f:
    long_desc = f.read()

setuptools.setup(
    name='testyoke',
    version=__version__,
    author='trevor grayson',
    author_email='trevor@ipsumllc.com',
    description='Provide insights on the outcomes of software test cases.',
    long_description=long_desc,
    long_description_content_type='text/markdown',
    url='https://github.com/trevorgrayson/testyoke',
    packages=setuptools.find_packages(),
    extras_require={
        'server': ["flask>=1.1.1"],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    python_requires='>=3.6',
)



#
# provide analytics on shas
#
#
from collections import defaultdict
from testyoke.models import ProjectState


PASS, FAIL = 'pass', 'fail'

class Analyzer: # inherits from GenericAnalyzer
    def __init__(self):
        self.shas = defaultdict(dict)
        self.real_shas = defaultdict(ProjectState)

    def analyze(self, case, suite=None, **options):
        # this should not be counting all tests.
        status = FAIL if case.did_fail else PASS
        if not self.shas[case.sha].get(status):
            self.shas[case.sha][status] = 0

        self.shas[case.sha][status] += 1

    def all(self):
        return self.shas

    def get(self, sha):
        """ given `sha` """
        state = self.shas.get(sha, {})

        if state is None:
            return None

        state = ProjectState(**{**state, **{'sha': sha}})
        return state

    @property
    def has_sha(self, sha):
        return self.shas.get(sha) is not None

    # def save(self):
    # def load(class):




PASS, FAIL = 'pass', 'fail'
TESTS_NAME = 'tests'

class TestSuite:
    def __init__(self, **kwargs):
        self.name = kwargs.get('name')

        self.cases = kwargs.get('cases', [])

        self.errors = int(kwargs.get('errors', 0))
        self.failures = int(kwargs.get('failures', 0))
        self.skipped  = int(kwargs.get('skipped', 0))
        self.tests = int(kwargs.get('tests', 0))
        self.time = float(kwargs.get('', 0.0))

    @property
    def to_dict(self):
        return {
            'name': self.name,
            'time': self.time
        }


class TestCase:
    def __init__(self, **kwargs):
        self.name = kwargs.get('name')
        self.classname = kwargs.get('classname')
        self.file = kwargs.get('file')
        self.line = kwargs.get('line')
        self.time = kwargs.get('time')
        self.latest_pass = kwargs.get('latest_pass')
        self.sha = kwargs.get('sha')

        # TODO get message "./failure/@message"
        self.message = kwargs.get('message')
        self.trace = kwargs.get('trace')

        self.line = kwargs.get('line')
        if self.line is not None:
            self.line = int(self.line)

        self.is_flaky = kwargs.get('flaky')
        self.is_regressive = kwargs.get('regressive')

    @property
    def ident(self):
        return self.name.lower().replace(" ", "_")

    @property
    def did_fail(self):
        return self.message is not None

    @property
    def to_dict(self):
        return {
            'name': self.name,
            'file': self.file,
            'classname': self.classname,
            'line': self.line,
            'time': self.time,
            'message': self.message,
            'trace': self.trace,
        }


class TestReport:
    def __init__(self, **kwargs):
        self.subject = kwargs.get('case')
        self.cases = kwargs.get('cases')
        self.passes = kwargs.get('passes')  # number
        self.fails = kwargs.get('fails')  # list of failures. compare exceptions?
        self.latest = kwargs.get('latest')

    @property
    def to_dict(self):
        return {}

    @property
    def pass_rate(self):
        fails = float(len(self.fails))
        passes = float(self.passes)

        return (passes-fails)/passes

    
    # def is_bullshit
    # def is_recurring
    # def is_flaky

    # def depends_on_service

    def to_dict(self):
        return {
            'pass_rate': self.pass_rate,
            # 'latest': {
            #     'timestamp': self.latest.timestamp,
            #     'sha': self.latest.sha
            # }
        }


class ProjectState:
    def __init__(self, **kwargs):
        self.sha = kwargs.get('sha')
        # self.status = kwargs.get('status')
        self.passes = kwargs.get('passes', kwargs.get(PASS, 
            kwargs.get(TESTS_NAME, {}).get('passes', 0)))
        self.fails = kwargs.get('fails', kwargs.get(FAIL, 
            kwargs.get(TESTS_NAME, {}).get('fails', 0)))

        self.failed = self.fails > 0

    @property
    def clean(self):
        """ Has never failed """
        return self.passes > 0 and self.fails == 0

    @property
    def flaky(self):
        """ Has both passed and failed """
        return self.passes > 0 and self.fails > 0

    @property
    def broken(self):
        """ Only failed """
        return self.passes == 0 and self.fails > 0

    @property
    def classification(self):
        if self.clean:
            return 'clean'
        elif self.flaky:
            return 'flaky'
        elif self.broken:
            return 'broken'
        else:
            return 'untested'

    @property
    def to_dict(self):
        return {
            'sha': self.sha, 
            TESTS_NAME: {
                'passes': self.passes,
                'fails': self.fails
            },
            'failed': self.failed,
            'nature': self.classification
        }

    def __repr__(self):
        return "nature: %s.  fails %s, passes %s" % (
            self.classification, self.fails, self.passes
        )
